#if defined(USE_LAPLACE)
#  include <df1b2fun.h>
#else
#  include <admodel.h>
#endif
//#include <parallel.h>
#include <signal.h>

void vm_initialize(void);
int have_jvm=0;

 //char cc[40]={"Copyright (c) 1993,1994,2000,2005 Otter Research Ltd"};

/*
#include "mpi.h"
class admpi_manager
{
  static const int MAX_MPI_OFFSET;
  MPI_Request * global_request;
  int * mpi_int;
  int num_slaves;
  int mpi_offset;
  MPI_Comm parent; 
  MPI_Comm everyone; 
  char worker_program[152];
  int master_flag;
  int do_minimize;
  int do_hess;
public:
  admpi_manager(int m,int argc,char * argv[]);
  admpi_manager::~admpi_manager();
  int is_master(void);
  void send_int_to_slave(int i,int slave_number);
  void increment_mpi_offset(void);
  int is_slave(void);
  void send_dvector_to_slave(const dvector& v,int slave_number);
  void get_int_from_master(int &i);
  void get_int_from_slave(int &i,int slave_number);
};
*/

#if defined(USE_ADMPI)
const int admpi_manager::MAX_MPI_OFFSET=1000;
dvector admpi_manager::get_dvector_from_master(void)
{
  MPI_Status status;
  int mmin,mmax;
  MPI_Recv(&mmin,1, MPI_INT,MPI_ANY_SOURCE,MPI_ANY_TAG,parent,&status);
  MPI_Recv(&mmax,1, MPI_INT,MPI_ANY_SOURCE,MPI_ANY_TAG,parent,&status);
  dvector tmp(mmin,mmax);
  int size=mmax-mmin+1;
  MPI_Recv(&(tmp(mmin)),size, MPI_DOUBLE,MPI_ANY_SOURCE,MPI_ANY_TAG,parent,&status);
  return tmp;
}

ivector admpi_manager::get_ivector_from_master(void)
{
  MPI_Status status;
  int mmin,mmax;
  MPI_Recv(&mmin,1, MPI_INT,MPI_ANY_SOURCE,MPI_ANY_TAG,parent,&status);
  MPI_Recv(&mmax,1, MPI_INT,MPI_ANY_SOURCE,MPI_ANY_TAG,parent,&status);
  ivector tmp(mmin,mmax);
  int size=mmax-mmin+1;
  MPI_Recv(&(tmp(mmin)),size,MPI_INT,MPI_ANY_SOURCE,MPI_ANY_TAG,parent,&status);
  cout << "SCA" << tmp << endl;
  return tmp;
}

void admpi_manager::send_slave_hessian_assignments(int nvar)
{
  int m=nvar/(1+num_hess_slaves);
  int excess=nvar-m*(1+num_hess_slaves);
  if (allocated(all_hess_bounds)) all_hess_bounds.deallocate();
  all_hess_bounds.allocate(0,num_hess_slaves,1,2);
  all_hess_bounds(0,1)=1;
  all_hess_bounds(0,2)=m;
  if (allocated(hess_bounds)) hess_bounds.deallocate();
  hess_bounds.allocate(1,2);
  hess_bounds=all_hess_bounds(0);
  for (int i=1;i<=num_hess_slaves;i++)
  {
    all_hess_bounds(i,1)=all_hess_bounds(i-1,2)+1;
    all_hess_bounds(i,2)=all_hess_bounds(i,1)+m-1;
    if (excess>0) 
    {
      all_hess_bounds(i,2)+=1;
      excess--;
    }
    cout << "In send slave hess ass from master" << endl;
    cout << "sending " << hess_bounds << endl;
    send_ivector_to_slave(all_hess_bounds(i),i);
    cout << "sent" << endl;
  }
}

void admpi_manager::get_slave_hessian_assignments(void)
{
  cout << "In get slave hess bounds from master" << endl;
  hess_bounds=get_ivector_from_master();
  cout << "got hess_bounds = " << hess_bounds << endl;
}

void admpi_manager::get_int_from_slave(int &i,int _slave_number)
{
  int slave_number=_slave_number-1;
  MPI_Status status;
  MPI_Recv(&i,1,MPI_INT,slave_number,0,everyone,&status);
}
void admpi_manager::get_int_from_master(int &i)
{
  MPI_Status status;
  MPI_Comm parent2; 
  MPI_Comm_get_parent(&parent2);
  // note this is a blocking receive
  cout << "SBA" << parent << endl;
  cout << "parent = " << parent << endl;
  MPI_Recv(&i,1, MPI_INT,MPI_ANY_SOURCE,MPI_ANY_TAG,parent2,&status);
  cout << "SBB" << parent << endl;
}

int admpi_manager::is_master(void)
{
  return (parent==MPI_COMM_NULL);
}

int admpi_manager::is_slave(void)
{
  return (parent!=MPI_COMM_NULL);
}
admpi_manager::~admpi_manager()
{
  delete [] global_request;
  global_request=0;
  delete [] mpi_int;
  mpi_int=0;
  MPI_Finalize();
}


void admpi_manager::send_dvector_to_slave(const dvector& v,
  int _slave_number)
{
  int slave_number=_slave_number-1;
  MPI_Status myStatus;
  /* make sure that the previous read using this memory area
     has completed 
  */
  if(global_request[mpi_offset])  /* checks to see if has been used at least */
                                  /* once */
  {
    MPI_Wait(&(global_request[mpi_offset]), &myStatus);
  }
  mpi_int[mpi_offset]=v.indexmin();
  MPI_Isend(&(mpi_int[mpi_offset]),1,MPI_INT,slave_number,0,
    everyone,&(global_request[mpi_offset]));
  increment_mpi_offset();
  if(global_request[mpi_offset])  /* checks to see if has been used at least */
                                  /* once */
  {
    MPI_Wait(&(global_request[mpi_offset]), &myStatus);
  }
  mpi_int[mpi_offset]=v.indexmax();

  MPI_Isend(&(mpi_int[mpi_offset]),1,MPI_INT,slave_number,0,
    everyone,&(global_request[mpi_offset]));
  increment_mpi_offset();
  int mmin=v.indexmin();
  int mmax=v.indexmax();
  int size=mmax-mmin+1;
  MPI_Request request;
  MPI_Isend(&(v[mmin]),size,MPI_DOUBLE,slave_number,0,everyone,&request);
}

void admpi_manager::send_ivector_to_slave(const ivector& v,int _slave_number)
{
  int slave_number=_slave_number-1;
  MPI_Status myStatus;
  /* make sure that the previous read using this memory area
     has completed 
  */
  if(global_request[mpi_offset])  /* checks to see if has been used at least */
                                  /* once */
  {
    MPI_Wait(&(global_request[mpi_offset]), &myStatus);
  }
  mpi_int[mpi_offset]=v.indexmin();
  MPI_Isend(&(mpi_int[mpi_offset]),1,MPI_INT,slave_number,0,
    everyone,&(global_request[mpi_offset]));
  increment_mpi_offset();
  if(global_request[mpi_offset])  /* checks to see if has been used at least */
                                  /* once */
  {
    MPI_Wait(&(global_request[mpi_offset]), &myStatus);
  }
  mpi_int[mpi_offset]=v.indexmax();

  MPI_Isend(&(mpi_int[mpi_offset]),1,MPI_INT,slave_number,0,
    everyone,&(global_request[mpi_offset]));
  increment_mpi_offset();
  int mmin=v.indexmin();
  int mmax=v.indexmax();
  int size=mmax-mmin+1;
  MPI_Request request;
  MPI_Isend(&(v[mmin]),size,MPI_INT,slave_number,0,everyone,&request);
  cout << "MBA" << endl;
  cout << "MCA" <<  v << endl;
  MPI_Wait(&(request), &myStatus);
  cout << "MBB" << endl;
}

void admpi_manager::send_int_to_slave(int i,int _slave_number)
{
  int slave_number=_slave_number-1;
  MPI_Status myStatus;
  MPI_Request request;
  /* make sure that the previous read using this memory area
     has completed 
  */
  if(global_request[mpi_offset])  /* checks to see if has been used at least */
                                  /* once */
  {
    MPI_Wait(&(global_request[mpi_offset]), &myStatus);
  }
  mpi_int[mpi_offset]=i;
  MPI_Isend(&(mpi_int[mpi_offset]),1,MPI_INT,slave_number,0,
    everyone,&(global_request[mpi_offset]));
  //sleep(1);
  MPI_Wait(&(global_request[mpi_offset]), &myStatus);
  increment_mpi_offset();
}

void admpi_manager::send_int_to_master(int i)
{
  MPI_Status myStatus;
  MPI_Request request;
  /* make sure that the previous read using this memory area
     has completed 
  */
  if(global_request[mpi_offset])  /* checks to see if has been used at least */
                                  /* once */
  {
    MPI_Wait(&(global_request[mpi_offset]), &myStatus);
  }
  mpi_int[mpi_offset]=i;
  MPI_Isend(&(mpi_int[mpi_offset]),1,MPI_INT,0,0,
    parent,&(global_request[mpi_offset]));
  //sleep(1);
  MPI_Wait(&(global_request[mpi_offset]), &myStatus);
  increment_mpi_offset();
}

void admpi_manager::increment_mpi_offset(void)
{
  mpi_offset++;
  if (mpi_offset==MAX_MPI_OFFSET)
  {
    mpi_offset=0;
  }
}

admpi_manager::admpi_manager(int m,int argc,char * argv[])
{
  parent=0;
  global_request = new MPI_Request[MAX_MPI_OFFSET];
  mpi_offset=0;
  num_slaves=0;
  num_hess_slaves=0;
  do_minimize=0;
  do_hess=0;
  mpi_int = new int[MAX_MPI_OFFSET];
 
  for (int i=0;i<MAX_MPI_OFFSET;i++)
  {
    global_request[i]=0;
    //mpi_int[i]=0;
  }
 
  int rank,world_size, universe_size, *universe_sizep, flag;
  if (strlen(argv[0])>150)
  {
    cerr << "program name too long for mpi manager" << endl;
    ad_exit(1);
  }
  else
  {
    strncpy(worker_program,argv[0],151);
  }

  MPI_Init(&argc, &argv);
  MPI_Comm_size(MPI_COMM_WORLD, &world_size);

  cout << world_size << endl;

  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  cout << "master rank " << rank << endl;
  MPI_Comm_get_parent(&parent);

  //sleep(1);
  if (is_master())
  {
    cout << "I am the master process" << endl;
    do_minimize=1;
    do_hess=1;
    if (world_size != 1) 
    {
      cerr << "Error --can only have one MPI master process" << endl;
      ad_exit(1);
    }

    //sleep(1);

    // !!! need to pass this number in MPI options
    int on=0; int nopt=0;
    if ( (on=option_match(argc,argv,"-nslaves",nopt))>-1)
    {
      if (nopt ==1)	    
      {	      
        num_slaves=atoi(argv[on+1]);
        num_hess_slaves=num_slaves;
      }
      else
      {
        cerr << "Wrong number of options to -nslaves -- must be 1"
          " you have " << nopt << endl;		
        ad_exit(1);
      }	
    }
    else
    {
      num_slaves=1;
      num_hess_slaves=1;
    }
   
    if (num_slaves < 1) 
    {
      cerr << "Error --no room for slaves" << endl;
      ad_exit(1);
    }

    //sleep(1);

    /*
     * Now spawn the slaves. Note that there is a run-time determination
     * of what type of slaves to spawn, and presumably this calculation must

     * be done at run time and cannot be calculated before starting
     * the program. If everything is known when the application is
     * first started, it is generally better to start them all at once
     * in a single MPI_COMM_WORLD.
     */
  
     fprintf(stderr, "will spawn %d slaves\n", num_slaves);
     cout << "there are " << num_hess_slaves << " hess slaves" << endl;
     int * ia=(int*)malloc(sizeof(int));
     char *myargvs[] = {"-slave", NULL};
     MPI_Comm_spawn(worker_program, myargvs, num_slaves,
            MPI_INFO_NULL, 0, MPI_COMM_SELF, &everyone,
            MPI_ERRCODES_IGNORE);
     //cout << "enter 2" << endl;
     //cin >> ia[0];
     ia[0]=2;
     // send an int to the slaves so that they will wait
     // until they get it
     for (int i=1;i<=num_slaves;i++)
     {
       send_int_to_slave(ia[0]+i,i);
     }
     //sleep(1);
  }
  if (is_slave())
  {
    cout << "eat shit I am a slave process" << endl;

    slave_number=rank+1;
    cout << "slave_number = " << slave_number << endl;
    do_minimize=0;
    do_hess=1;
    int j=-1;
    //cin >> j;
    int i=-1;
    cout << "SAA" << endl;
    get_int_from_master(i);
    cout << "SBB" << endl;
    cout << "In slave i = " << i << endl;
    //sleep(2);
  }
}

#endif

void  strip_full_path(BOR_CONST adstring& _s)
{
  adstring& s = (adstring&) _s; 
  int n=s.size();
  int i=0;
  for (i=n-1;i>=1;i--)
  {
    if ( s(i)=='\\' || s(i) == '/' || s(i)==':') break;
  }
  s=s(i+1,n);
}  

void set_signal_handlers(void)
{
  signal(SIGFPE,exit_handler);
  signal(SIGSEGV,exit_handler);
  signal(SIGILL,exit_handler);
  signal(SIGINT,exit_handler);
}



ad_comm::ad_comm(int _argc,char * _argv[])
{
  ad_comm::argc=_argc;
  ad_comm::argv=_argv;
  int pvm_flag=0;
  if (option_match(_argc,_argv,"-time")>-1)  
  {
    time_flag=1;
  }
  else
  {
    time_flag=0;
  }
  if (time_flag)
  {
    if (!ptm)
    {
      ptm=new adtimer();
    }
    if (!ptm1)
    {
      ptm1=new adtimer();
    }
  }
  no_atlas_flag=0;
  if (option_match(_argc,_argv,"-noatlas")>-1)  no_atlas_flag=1;
  // use pvm_flag for both PVM and MPI
  if (option_match(_argc,_argv,"-slave")>-1)  pvm_flag=2;
  if (option_match(_argc,_argv,"-master")>-1) pvm_flag=1;

#if defined(USE_ADMPI)
  if (pvm_flag)
    mpi_manager = new admpi_manager(pvm_flag,_argc,_argv);
  else
#endif
    mpi_manager = NULL;

#if defined(USE_ADPVM)
  if (pvm_flag)
    pvm_manager = new adpvm_manager(pvm_flag);
  else
#endif
    pvm_manager = NULL;

  
#if defined(USE_ADPVM)
  if (pvm_manager)
  {
    if (pvm_manager->mode==2)   //slave
    {
      int on=0; int nopt=0;
      if ( (on=option_match(_argc,_argv,"-slave",nopt))>-1)
      {
        if (nopt ==1)	    
        {	      
          pvm_manager->slave_number=atoi(ad_comm::argv[on+1]);
        }
        else
        {
          cerr << "Wrong number of options to -slave -- must be 1"
            " you have " << nopt << endl;		
          ad_exit(1);
        }	
      }   
      if ( (on=option_match(_argc,_argv,"-slavedir",nopt))>-1)
      {
        if (nopt ==1)	    
        {	      
          ad_chdir(_argv[on+1]);
        }
        else
        {
          cerr << "Wrong number of options to -slavedir -- must be 1"
            " you have " << nopt << endl;		
        }	
      }   
    }
  }
#endif

  /*
  if (option_match(_argc,_argv,"-gui")>-1)
  {
    vm_initialize();
  }
  */
  set_signal_handlers();
  adprogram_name=_argv[0];
  int len=strlen(_argv[0]);
  for (int i=1;i<=len;i++) adprogram_name[i]=tolower(adprogram_name[i]);
#if !defined(__SPDLL__)
  strip_full_path(adprogram_name);
#endif
  adstring workdir;
  ad_getcd(workdir);
  if (_argc>1)
  {
    if (option_match(_argc,_argv,"-?")>-1
      || option_match(_argc,_argv,"-help")>-1 )
    {
      // remove the first part of the path if present
      int i;
      for (i=adprogram_name.size();i>=1;i--)
      {
        if (adprogram_name(i)=='\\')
        {
          adprogram_name=adprogram_name(i+1,adprogram_name.size());
          break;
        }
      }

      for (i=1;i<=adprogram_name.size();i++)
      {
        if (adprogram_name(i)=='.') break;
      }
      int n=i-1;
      adprogram_name=adprogram_name(1,n);
      (*ad_printf)( " AD Model Builder version 8.0.2"
              " copyright (c) 1993 2002 2005 2007 Otter Research Ltd\n");
      (*ad_printf)(" USAGE--%s options\n",(char*)(adprogram_name));
      (*ad_printf)(" where an option consists of -option_tag followed by arguments if necessary\n");
      (*ad_printf)( " -ainp NAME      change default ascii input parameter file name to NAME\n");
      (*ad_printf)( " -binp NAME      change default binary input parameter file name to NAME\n");
      (*ad_printf)( " -est            only do the parameter estimation\n");
      (*ad_printf)( " -noest          do not do the parameter estimation (optimization) \n");
      (*ad_printf)( " -ind NAME       change default input data file name to NAME\n");
      (*ad_printf)( " -lmn N          use limited memory quasi newton -- keep N steps\n");
      (*ad_printf)( " -dd N           check derivatives after n function evaluations\n");
      (*ad_printf)( " -lprof          perform profile likelihood calculations\n");
      (*ad_printf)( " -maxph N        increase the maximum phase number to N\n");
      (*ad_printf)( " -mcdiag         use diagonal covariance matrix for mcmc with diagonal values 1\n");
      (*ad_printf)( " -mcmc [N]       perform markov chain monte carlo with N simulations\n");
      (*ad_printf)( " -mcmult N       multiplier N for mcmc default\n");
      (*ad_printf)( " -mcr            resume previous mcmc\n");
      (*ad_printf)( " -mcrb  N        reduce the amount of correlation in the covariance matrix 1<=N<=9\n");
      (*ad_printf)( " -mcnoscale      don't rescale step size for mcmc depending on acceptance rate\n");
      (*ad_printf)( " -nosdmcmc       turn off mcmc histogram calcs to make mcsave run faster\n");
      (*ad_printf)( " -mcgrope N      use probing strategy for mcmc with factor N\n");
      (*ad_printf)( " -mcseed N       seed for random number generator for markov chain monte carlo\n");
      (*ad_printf)( " -mcscale N       rescale step size for first N evaluations\n");
      (*ad_printf)( " -mcsave N       save the parameters for every N'th simulation\n");
      (*ad_printf)( " -mceval         Go throught the saved mcmc values from a previous mcsave\n");
//      (*ad_printf)( " -mcu            use uniformaly distributed steps for mcmc\n");
      (*ad_printf)( " -crit N1,N2,... set gradient magnitude convergence criterion to N\n");
      (*ad_printf)( " -iprint N       print out function minimizer report every N iterations\n");
      (*ad_printf)( " -maxfn N1,N2,.. set maximum number opf function eval's to N\n");
      (*ad_printf)( " -rs             if function minmimizer can't make progress rescale and try again\n");
  //    (*ad_printf)( " -sp             for DLL running from splus write to command window\n");
      (*ad_printf)( " -nox            don't show vector and gradient values in function minimizer screen report\n");
      (*ad_printf)( " -phase N        start minimization in phase N\n");
      (*ad_printf)( " -simplex        use simplex algorithm for minimization (new test version)\n");
      (*ad_printf)( " -nohess         don't do hessian or delta method for std dev\n");
      (*ad_printf)( " -eigvec         calculate eigenvectors of the Hessian\n");
      (*ad_printf)( " -sdonly         do delta method for std dev estimates without redoing hessian\n");
      (*ad_printf)( " -ams N          set arrmblsize to N (ARRAY_MEMBLOCK_SIZE)\n");
      (*ad_printf)( " -cbs N          set CMPDIF_BUFFER_SIZE to N (ARRAY_MEMBLOCK_SIZE)\n");
      (*ad_printf)( " -mno N          set the maximum number of independent variables to N\n");
      (*ad_printf)( " -mdl N          set the maximum number of dvariables to N\n");
      (*ad_printf)( " -master         run as PVM master program\n");
      (*ad_printf)( " -gbs N          set GRADSTACK_BUFFER_SIZE to N (ARRAY_MEMBLOCK_SIZE)\n");
      (*ad_printf)( " -master         run as PVM master program\n");
      (*ad_printf)( " -slave          run as PVM slave program\n");
      (*ad_printf)( " -pvmtime        record timing information for PVM performance analysis\n");
      (*ad_printf)( " -? or -help     this message\n");
      (*ad_printf)( " Random effects options if applicable\n");
    //if (function_minimizer::random_effects_flag)
    {
      (*ad_printf)( " -nr N           maximum number of Newton-Raphson steps\n");       
      (*ad_printf)( " -imaxfn N       maximum number of fevals in quasi-Newton inner optimization\n");       
      (*ad_printf)( " -is N  [M]      set importance sampling size to n for random effects\n                    if present M is used as RN seed\n");       
      (*ad_printf)( " -isf N          set importance sampling size funnel blocksto n for random effects\n");       
      (*ad_printf)( " -isdiag         print importance sampling diagnostics\n");
      (*ad_printf)( " -hybrid         do hybrid Monte Carlo version of MCMC\n");
      (*ad_printf)( " -hbf            set the hybrid bounded flag for bounded parameters\n");
      (*ad_printf)( " -hyeps          mean step size for hybrid Monte Carlo\n");
      (*ad_printf)( " -hynstep        number of steps for hybrid Monte Carlo\n");
      (*ad_printf)( " -noinit         do not initialize random effects before"
           " inner optimzation\n");       
      (*ad_printf)( " -ndi N          set maximum number of separable calls\n");       
      (*ad_printf)( " -noreders       turn off re's for derivative calcs for std devs\n");       
      (*ad_printf)( " -ndb N          set number of blocks for derivatives for"
           " random effects (reduces temporary file sizes)\n");       
      (*ad_printf)( " -ndv N          set the maximum number of dependent"
           " variables\n");       
      (*ad_printf)( " -ddnr           use high precision Newton-Raphson for"
           " inner optimization for banded hessian case ONLY even if implemented\n");       
      (*ad_printf)( " -nrdbg           verbose reporting for debugging newton-raphson\n");       
#  if defined(__MINI_MAX__)
      (*ad_printf)( " -mm N          do minimax optimization\n");
#  endif
      (*ad_printf)( " -shess         use sparse Hessian structure"
           " inner optimzation\n");       
    }
      ad_exit(0);
    }
  }
  allocate();
}

ad_comm::ad_comm(void)
{
  allocate();
}

void ad_comm::allocate(void)
{
#if defined (_WIN32)
 directory_prefix='\\';
#else
 directory_prefix='/';
#endif
  adstring tmpstring;
  // remove the first part of the path if present
  int i;
#if !defined(__SPDLL__)
  for (i=adprogram_name.size();i>=1;i--)
  {
    if (adprogram_name(i)==directory_prefix)
    {
      adprogram_name=adprogram_name(i+1,adprogram_name.size());
      break;
    }
  }

#endif
  int n=adprogram_name.size();

  // strip off the full path to get just the file name
  int n1=1;
  for (i=adprogram_name.size()-1;i>1;i--)
  {
    if ( adprogram_name(i)== directory_prefix )
    {
      n1=i+1;
      break;
    }
  }

  // strip off the .exe if it is there
  for (i=n1+1;i<=adprogram_name.size()-1;i++)
  {
    if ( adprogram_name(i)=='.' 
      && adprogram_name(i+1) != directory_prefix )
    {
      n=i-1;
      break;
    }
  }
  adprogram_name=adprogram_name(n1,n);
  // change the working directory name
  if (argc > 1)
  {
    int on=0;
    if ( (on=option_match(argc,argv,"-wd"))>-1)
    {
      if (on>argc-2 || argv[on+1][0] == '-') 
      { 
        cerr << "Invalid input data command line option"
           " -- ignored" << endl;  
      }
      else
      {
        tmpstring = adstring(argv[on+1]);
	wd_flag=1;
      }
    }
  }
  if (length(tmpstring))
  {
    if (tmpstring(length(tmpstring)) == directory_prefix)
    {
      adprogram_name=tmpstring + adprogram_name;
      working_directory_path = tmpstring;
    }
    else
    {
      adprogram_name=tmpstring + directory_prefix + adprogram_name;
      working_directory_path = tmpstring + directory_prefix;
    }
  }
  
  tmpstring=adprogram_name + adstring(".dat");
  if (argc > 1)
  {
    int on=0;
    if ( (on=option_match(argc,argv,"-ind"))>-1)
    {
      if (on>argc-2 || argv[on+1][0] == '-') 
      { 
        cerr << "Invalid input data command line option"
           " -- ignored" << endl;  
      }
      else
      {
        tmpstring = adstring(argv[on+1]);
      }
    }
  }
  global_datafile= new cifstream(tmpstring);
  if (!global_datafile)
  {
    cerr << "Error trying to open data input file "
         <<  tmpstring << endl;
  }
  else
  {
    if (!(*global_datafile))
    {
      cerr << "Error trying to open data input file "
         <<  tmpstring << endl;
      delete global_datafile;
      global_datafile=NULL;
    }
  }
  adstring ts=adprogram_name + adstring(".log");
  global_logfile= new ofstream( (char*)ts);

  int biopt=-1;
  int aiopt=-1;
  biopt=option_match(argc,argv,"-binp");
  aiopt=option_match(argc,argv,"-ainp");

  tmpstring=adprogram_name + adstring(".bin");
  if (!global_bparfile && aiopt == -1)
  {
    if (biopt>-1)
    {
      if (biopt>argc-2 || argv[biopt+1][0] == '-')
      { 
        cerr << "Invalid input parameter file command line option"
           " -- ignored" << endl;  
      }
      else
      {
        tmpstring = adstring(argv[biopt+1]);
      }
    }
    global_bparfile= new uistream(tmpstring);
    if (global_bparfile)
    {
      if (!(*global_bparfile))
      {
	if (biopt>-1)
	{
	  cerr << "Error trying to open binary inoput par file "
	       << tmpstring << endl;
	  exit(1);
	}
        delete global_bparfile;
        global_bparfile=NULL;
      }
    }
  }
  tmpstring=adprogram_name + adstring(".pin");
  if (!global_parfile)
  {
    if (aiopt>-1)
    {
      if (aiopt>argc-2 || argv[aiopt+1][0] == '-')
      { 
        cerr << "Invalid input parameter file command line option"
           " -- ignored" << endl;  
      }
      else
      {
        tmpstring = adstring(argv[aiopt+1]);
      }
    }
    global_parfile= new cifstream(tmpstring);
    if (global_parfile)
    {
      if (!(*global_parfile))
      {
	if (aiopt>-1)
	{
	  cerr << "Error trying to open ascii inoput par file "
	       << tmpstring << endl;
	  exit(1);
	}
        delete global_parfile;
        global_parfile=NULL;
      }
    }
  }
}

ad_comm::~ad_comm()
{
  if (mpi_manager)
  {
    delete mpi_manager;
    mpi_manager=0;
  }

  if (ptm)
  {
    delete ptm;
    ptm=0;
  }
  if (ptm1)
  {
    delete ptm1;
    ptm1=0;
  }
  if (global_datafile)
  {
    delete global_datafile;
    global_datafile=NULL;
  }
  if (global_parfile)
  {
    delete global_parfile;
    global_parfile=NULL;
  }
  if (global_logfile)
  {
    delete global_logfile;
    global_logfile=NULL;
  }
}

void function_minimizer::pre_userfunction(void)
{
#if defined(USE_LAPLACE)
  if (lapprox)
  {
    if (lapprox->hesstype==2) 
    {
      //lapprox->num_separable_calls=0;
      lapprox->separable_calls_counter=0;
    }
  }
#endif
  userfunction();
#if defined(USE_LAPLACE)
  int i;
  if (lapprox)
  {
    if (lapprox->hesstype==2) 
    {
      lapprox->num_separable_calls=lapprox->separable_calls_counter;

      double tmp=0.0;
      int inner_opt_value=inner_opt();
      if (lapprox->saddlepointflag==2)
      {
        if (inner_opt_value !=0 )
        {
          for (i=1;i<=lapprox->num_separable_calls;i++)
          {
            tmp-=(*lapprox->separable_function_difference)(i);
          }
          value(*objective_function_value::pobjfun)=tmp;
        }
      }
      else
      {
        for (i=1;i<=lapprox->num_separable_calls;i++)
        {
          tmp+=(*lapprox->separable_function_difference)(i);
        }
        value(*objective_function_value::pobjfun)=tmp;
      }
    }
  }
#endif
}

